#include <routing/address.h>
#include <trace/cmu-trace.h>
#include <common/scheduler.h>

#include <stdarg.h> // for logger

#include <vector>

#include "dtnmc.h"
#include "dtnmc_pkt.h"

int hdr_dtnmc_pkt::offset_;
int hdr_sumvec_pkt::offset_;

static class DtnmcHeaderClass : public PacketHeaderClass {
   public:
     DtnmcHeaderClass() : PacketHeaderClass("PacketHeader/Dtnmc", sizeof(hdr_dtnmc_pkt)){
        bind_offset(&hdr_dtnmc_pkt::offset_);
        bind_offset(&hdr_sumvec_pkt::offset_);
     }
} class_rtProtoDtnmc_hdr;

static class DtnmcClass : public TclClass
{
public:
   DtnmcClass() : TclClass("Agent/Dtnmc"){}

   TclObject* create(int argc, const char*const*argv) {
      assert(argc == 5);
      return (new Dtnmc((nsaddr_t)Address::instance().str2addr(argv[4])));
   }
} class_rtProtoDtnmc;

void Dtnmc_PktTimer::expire(Event *e) {
   if ( agent_->nstat() ) { // static nodes just listen hello messages and does nothing else.
      return;
   }

   /* If a node sends a packet near the end of simulation and simulation ends before it gets delivered,
    * the simulation results will be sckewed. So we use a stop_time field to mark when the nodes should
    * sending new messages. Preferably, this value should be equal to simulation time minus the TTL.
    */
   if ( CURRENT_TIME > agent_->stop_time() ) {
      fprintf(stderr, "LOG: Current time: %g, stop_time: %g\n", CURRENT_TIME, agent_->stop_time());
      return;
   }

   if ( agent_->cmode() == UNICAST ) {
      agent_->send_dtnmc_pkt();
   } else if ( agent_->cmode() == MULTICAST ) {
      agent_->send_dtnmc_mcast_pkt();
   }

   agent_->reset_pkt_timer();
}

void Dtnmc_HelloTimer::expire(Event *e) {
   if ( agent_->nstat() ) { // static nodes just listen hello messages and does nothing else.
      return;
   }

   agent_->sendHello();
   agent_->reset_hello_timer();
}

Dtnmc::Dtnmc(nsaddr_t id) : Agent(PT_DTNMC), pkt_timer_(this), hello_timer_(this) {
   bind("spray_", &spray_);    // spray-and-wait type
   bind("nocopy_", &nocopy_);  // number of copies to be spread
   bind("num_nodes_", &num_nodes_); // number of nodes in the system
   bind("core_", &core_);  // if the node is core node
   bind("nstat_", &nstat_); // if the node is static
   bind("xpos_", &xpos_);  // x-coordinate of node
   bind("ypos_", &ypos_);  // y-coordinate of node
   bind("stop_time_", &stop_time_); // time at which nodes stop generating messages
   bind("msg_ttk_", &msg_ttk_);     // time to kill (message expiration time)
   bind("msg_ttl_", &msg_ttl_);     // maxium number of hops (TTL)
   bind("mcast_base_addr", &mcast_base_addr_); // multicast group id
   bind("intra_group_", &intra_group_);        // should message exchange occur within the group only 
   bind("inter_group_prob_", &inter_group_prob_); // probability of message exchange among different groups

   ra_addr_ = id;   // router agent address
   seq_num_ = 0;    // message sequence number
   rtable_.id() = id;   // let the router table know the agent address
   cmode() = UNICAST;   // communication mode: unicast or multicast
   brcount = 0;         // buffer report counter

   // if spray is 2, it means to use source spray and wait, rather than binary spray and wait.
   if ( spray_ == 2 ) {
      rtable_.source_spray() = true;
   } else if ( spray_ == 3 ) { // 3 for epidemic routing
      rtable_.epidemic() = true;
   }

   // if the node is core, increase the buffer size
   if ( core_ == 1 ) {
      rtable_.set_capacity(CORE_BUF_SIZE);
   }

   log("LOG: M _%d_ [%d] [%d] %g (%g) %g\n", id, mcast_base_addr_, core_, CURRENT_TIME, stop_time(), msg_ttk_);
}

int Dtnmc::command(int argc, const char*const* argv)
{
   if ( argc == 2 ) {
      if(strcasecmp(argv[1], "start") == 0) {
         pkt_timer_.resched(0.0);
         //hello_timer_.resched(0.0);
         hello_timer_.resched((double) (HELLO_INTERVAL*Random::uniform()));
         return TCL_OK;
      } else if ( strcasecmp(argv[1], "print_rtable") == 0 ) {
         if ( logtarget_ != 0 ) {
            sprintf(logtarget_->pt_->buffer(), "P %f _%d_ Routing Table",
              CURRENT_TIME, ra_addr());
            logtarget_->pt_->dump();
            rtable_.print(logtarget_);
         } else { // no logtarget_ 
            fprintf(stdout, "%f _%d_ If you want to print this routing table "
                            "you must create a trace file in your tcl script",
                            CURRENT_TIME, ra_addr());
         }
         return TCL_OK;
      } else if ( strcasecmp(argv[1], "greet") == 0 ) {
         fprintf ( stdout, "Hello!\n"); // test feature
         return TCL_OK;
      }
   } else if ( argc == 3 ) {
      // obtains corresponding dmux to carry packets to upper layers
      if ( strcmp(argv[1], "port-dmux") == 0 ) {
         dmux_ = (PortClassifier*) TclObject::lookup(argv[2]);

         if ( dmux_ == 0 ) {
            fprintf(stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1], argv[2]);
            return TCL_ERROR;
         }
         return TCL_OK;
      } else if (strcmp(argv[1], "log-target")==0 || strcmp(argv[1], "tracetarget")==0) {
         logtarget_ = (Trace*) TclObject::lookup(argv[2]);
         if ( logtarget_ == 0 ) {
            return TCL_ERROR;
         }
         return TCL_OK;
      }
   }

   return Agent::command(argc, argv);
}

void Dtnmc::recv(Packet *p, Handler *h) {
   struct hdr_cmn *ch = HDR_CMN(p);
   struct hdr_ip *ih = HDR_IP(p);
   struct hdr_dtnmc_pkt *ph = HDR_DTNMC_PKT(p);

   if ( ih->saddr() == ra_addr() ) {
      if ( ch->num_forwards() > 0 ) { // drop packet if a loop exists
         drop(p, DROP_RTR_ROUTE_LOOP);
         return;
      } else if ( ch->num_forwards() == 0 ) {
         // this is a packet I am originating, must add IP header
         ch->size() += IP_HDR_LEN; // I don't know why we have to do this.
      }
   }

   // static node only receives HELLO packets only
   if ( nstat() ) {
      if ( ph->dh_type() == DTNMCTYPE_HELLO ) {
         recv_hello_pkt(p);
      }
      return;
   }

   if ( ch->ptype() == PT_DTNMC ) { // this is a dtnmc packet, must process it
      ih->ttl() = ih->ttl()-1;

      if ( ih->ttl() < 0 ) {
         free(p);
         return;
      }

      switch ( ph->dh_type() ) {
         case DTNMCTYPE_PKT:       // unicast operations
            recv_dtnmc_pkt(p);
            break;
         case DTNMCTYPE_MCAST_PKT:       // multicast operations
            recv_dtnmc_mcast_pkt(p);
            break;
         case DTNMCTYPE_DELEGATE:
            recv_delegate_pkt(p);
            break;
         case DTNMCTYPE_HELLO:
            recv_hello_pkt(p);
            break;
         case DTNMCTYPE_SUMVEC:
            recv_sumvec_pkt(p);
            break;
         default:
            struct hdr_sumvec_pkt *svh = HDR_SUMVEC_PKT(p);
            if ( svh->dh_type() == DTNMCTYPE_SUMVEC ) {
               recv_sumvec_pkt(p);
               break;
            } else if ( svh->dh_type() == DTNMCTYPE_REQVEC ) {
               recv_reqvec_pkt(p);
               break;
            }
            fprintf (stderr, "invalid DTNMC type (%x)\n", ph->dh_type() );
            break;
      }
   } else {
      ih->ttl_--;
      if ( ih->ttl_ == 0 ) {
         drop(p, DROP_RTR_TTL);
         return;
      }
      forward_data(p);
   }
}

void Dtnmc::recv_dtnmc_pkt(Packet *p) {
   struct hdr_ip *ih        = HDR_IP(p);
   struct hdr_dtnmc_pkt *ph = HDR_DTNMC_PKT(p);

   // all routing messages are sent from and to port RT_PORT, check it
   assert(ih->sport() == RT_PORT);
   assert(ih->dport() == RT_PORT);

   assert(ph->dh_type() == DTNMCTYPE_PKT);

   if ( ra_addr() == ih->daddr() ) {
      if(!rtable_.lookup_received_entry(ph->pkt_osrc(), ph->pkt_seq_num())){ //never received this packet
         if ( ph->member_till() < CURRENT_TIME ) { // drop if membership time is over
            log("LOG: o _%d_ _%d_ [%d] [%d] %g [%g,%g]\n", ra_addr(), ph->pkt_osrc(), ph->mcast_grp_id(), ph->pkt_seq_num(), CURRENT_TIME, ph->member_since(), ph->member_till()) ;
         } else {
            rtable_.received_add_entry(ph->pkt_osrc(), ph->pkt_seq_num());
            rtable_.store_app_packet(p);
            log("LOG: r _%d_ _%d_ %d [%d] %g %g %d\n", ra_addr(), ph->pkt_osrc(), ph->pkt_src(), ph->pkt_seq_num(), ph->gen_time(), CURRENT_TIME, DTNMC_DEF_TTL - ih->ttl());
         }
      } else { // we have received this packet before, ignore (multiple copies).
         log("LOG: i _%d_ _%d_ [%d] %g\n", ra_addr(), ph->pkt_osrc(), ph->pkt_seq_num(), CURRENT_TIME);
      }
   }

   Packet::free(p);
   return;
}

void Dtnmc::recv_dtnmc_mcast_pkt(Packet *p) {
   struct hdr_ip *ih        = HDR_IP(p);
   struct hdr_dtnmc_pkt *ph = HDR_DTNMC_PKT(p);

   // all routing messages are sent from and to port RT_PORT, check it
   assert(ih->sport() == RT_PORT);
   assert(ih->dport() == RT_PORT);

   assert(ph->dh_type() == DTNMCTYPE_MCAST_PKT);

   // we are not in this group, drop packet
   // TODO: adjust this for multiple group membership scenarios
   if ( ph->mcast_grp_id() == mcast_base_addr() ) {
      if ( ra_addr() == ih->daddr() ) {
         if ( ph->member_till() < CURRENT_TIME ) { // drop if membership time is over
            log("LOG: o _%d_ _%d_ [%d] [%d] %g [%g,%g]\n", ra_addr(), ph->pkt_osrc(), ph->mcast_grp_id(), ph->pkt_seq_num(), CURRENT_TIME, ph->member_since(), ph->member_till()) ;
         } else { // packet is in membership interval
            if(!rtable_.lookup_received_entry(ph->pkt_osrc(), ph->pkt_seq_num())){ // never received this packet
               rtable_.received_add_entry(ph->pkt_osrc(), ph->pkt_seq_num());
               rtable_.store_app_packet(p);
               rtable_.store_packet(p);  // store this packet for other group members.
               log("LOG: R _%d_ _%d_ %d [%d] [%d] %g %g %d\n", ra_addr(), ph->pkt_osrc(), ph->pkt_src(), ph->pkt_seq_num(), ph->mcast_grp_id(), ph->gen_time(), CURRENT_TIME, DTNMC_DEF_TTL - ih->ttl()) ;
            } else { // we have received this packet before, ignore (multiple copies).
               log("LOG: I _%d_ _%d_ %d [%d] [%d] %g\n", ra_addr(), ph->pkt_osrc(), ph->pkt_src(), ph->pkt_seq_num(), ph->mcast_grp_id(), CURRENT_TIME) ;
               //exit(0);
            }
         }
      }
   }

   // Release resources
   Packet::free(p);
}

void Dtnmc::recv_delegate_pkt(Packet *p) {
   struct hdr_ip *ih        = HDR_IP(p);
   struct hdr_dtnmc_pkt *ph = HDR_DTNMC_PKT(p);

   // all routing messages are sent from and to port RT_PORT, check it
   assert(ih->sport() == RT_PORT);
   assert(ih->dport() == RT_PORT);

   if ( ra_addr() != ih->daddr() ) { // this packet is not for us
      Packet::free(p);
      return;
   }

   if ( ra_addr() == ph->pkt_dest() ) {
      log("ERROR: node: %d, orig: %d, from: %d, dest: %d, nc: %d\n", ra_addr(), ph->pkt_osrc(), ih->saddr(), ph->pkt_dest(), ph->no_copy());
   } else {
      if ( ra_addr() == ih->daddr() ) { 
         if(!rtable_.lookup_received_entry(ph->pkt_osrc(), ph->pkt_seq_num())) { //never received this packet
            rtable_.received_add_entry(ph->pkt_osrc(), ph->pkt_seq_num());
            //log("pdel: node: %d, orig: %d, from: %d, dest: %d, seqno: %d, gid: %d, nc: %d\n", ra_addr(), ph->pkt_osrc(), ih->saddr(), ph->pkt_dest(), ph->pkt_seq_num(), ph->mcast_grp_id(), ph->no_copy());
            rtable_.store_packet(p);
         }
      }
   }

   // Release resources
   Packet::free(p);
}

void Dtnmc::recv_hello_pkt(Packet *p) {
   struct hdr_ip *ih        = HDR_IP(p);
   struct hdr_dtnmc_pkt *ph = HDR_DTNMC_PKT(p);

   // all routing messages are sent from and to port RT_PORT, check it
   assert(ih->sport() == RT_PORT);
   assert(ih->dport() == RT_PORT);

   int sid = ih->saddr(); 
   int gid = ph->mcast_grp_id();

   double lastHeard = rtable_.last_hello(sid);
   double firstHeard = rtable_.first_hello(sid);

   // We log the first meeting time. To avoid loggging of a single meeting
   // by both of the nodes, only the one with lower id logs meeting time.
   if ( (CURRENT_TIME - lastHeard > HELLO_EXPIRATION_INTV) && (ra_addr() < sid || nstat())) {
      if ( lastHeard == DMIN ) {
          lastHeard = 0;
      } else if ( firstHeard != DMIN ) { // record how long the nodes stayed in contact range
         if ( nstat() ) {
            /* due to HELLO_INTERVAL granularity, we detect a node after 0.5*HELLO_INTERVAL seconds after
             * it enters radio range, and we record 0.5*HELLO_INTERVAL less when it exits. So we
             * make this up by adding 1.0*HELLO_INTERVAL to the neighboring time.
             */ 
            log("LOG: h _%d_ _%d_ [%d] [%d] %g %g\n", ra_addr(), sid, xpos(), ypos(), lastHeard - firstHeard, CURRENT_TIME);
         } else {
            log("LOG: N _%d_ _%d_ %g\n", ra_addr(), sid, lastHeard - firstHeard );
         }
      }

      // this is the first time we meet since the two nodes parted last
      rtable_.update_first_hello(sid, CURRENT_TIME);

      if ( nstat() ) {
         log("LOG: H _%d_ _%d_ [%d] [%d] %g %g\n", ra_addr(), sid, xpos(), ypos(), CURRENT_TIME - lastHeard, CURRENT_TIME);
      } else {
         log("LOG: Z _%d_ _%d_ %g\n", ra_addr(), sid, CURRENT_TIME - lastHeard);
      }
   }

   rtable_.update_hello(sid, CURRENT_TIME);

   if ( nstat() ) { // nothing to do for a info retrieving static node
      Packet::free(p);
      return;
   }

   rtable_.update_grp_hello(gid, CURRENT_TIME);

   //log("node %d (gid: %d): received hello from node %d at time: %g\n", ra_addr(), mcast_base_addr(), sid, CURRENT_TIME);

   std::vector <int> new_neighbors;
   for ( int i = 0; i < MAX_NEIGHBOR; i++ ) {
      nsaddr_t nbr = ph->neighbor(i);
      if ( nbr != NEIGHBOR_END && nbr != ra_addr() ) { // don't include self among neighbors
         new_neighbors.push_back(nbr);
         //log("node %d (gid: %d): %d has neighbor %d\n", ra_addr(), mcast_base_addr(), sid, nbr);
      }
   }
   rtable_.set_neighbors(sid, new_neighbors);

   // just heard a hello, send stored packets for that node
   if ( cmode() == UNICAST ) {
      send_stored_pkt(sid);
   } else if ( cmode() == MULTICAST ) {
      //send_stored_mcast_pkt(sid, gid);
   }

   // delegate some packets to this neighbour if it just came in range 
   //if ( CURRENT_TIME - lastHeard > HELLO_EXPIRATION_INTV ) {
   if ( spray_ && ! rtable_.epidemic() ) {
      if ( cmode() == UNICAST ) {
         delegate_pkt(sid);
      } else if ( cmode() == MULTICAST ) {
         delegate_mcast_pkt(sid, gid);
      }
   }
   //}

   // send summary vector of the packets that we have
   if ( rtable_.epidemic() ) {
      if ( rtable_.last_sumvecTime(sid) < (CURRENT_TIME - SUMVEC_INTV) ) {
         // if intra_group option is set, a node only communicates with its group members only 
         if ( cmode() == MULTICAST ) {
            if ( intra_group() ) { // intra_group is defined, we only exchange packets within group
               if ( mcast_base_addr() == gid ) {
                  send_sumvec_pkt(sid);
               }
            } else { 
               send_sumvec_pkt(sid);
            }
         } else { // UNICAST
            send_sumvec_pkt(sid);
         }
      }
   }

   //if ( !core_ && ph->is_core() ) { // this makes core nodes not exchange packets
   if ( ph->is_core() ) { // core nodes exchange packets
      delegate_to_core(sid);
   }

   Packet::free(p);
}

void Dtnmc::recv_reqvec_pkt(Packet *p)
{
   int i;
   int count = 0;
   int osrc;
   int seqno;
   int gid;
   bool should_send;
   struct dtnmc_pkt_info *pin;

   struct hdr_cmn *ch = HDR_CMN(p);
   struct hdr_ip *ih = HDR_IP(p);
   struct hdr_sumvec_pkt *ph = HDR_SUMVEC_PKT(p);

   gid = ph->mcast_grp_id();

   assert(ih->sport() == RT_PORT);
   assert(ih->dport() == RT_PORT);

   if ( ra_addr() != ph->pkt_dest() ) {
      Packet::free(p);
      return;
   }

   if ( ph->vsize() < 1 ) {
      Packet::free(p);
      return;
   }

   nsaddr_t sid = ph->pkt_src();
   nsaddr_t dest = sid; 

   while ( (pin = rtable_.pop_delegate_packet_info(dest)) != NULL ) {
      should_send = false; 

      if ( pin->dest == ra_addr() ) {
         free(pin);
         continue;
      }

      if ( gid != pin->gid ) {
         /* If messages are only exchanged within the group, we don't send messages to non-group members.
          * If the group destination group id is different from the group id of the current packet
          * we apply the inter-group message exchange probability to control the load of a message
          * on non-group members.
          */
         //if ( intra_group() || pin->gid != mcast_base_addr() || Random::uniform() >= inter_group_prob() ) { 
         if ( intra_group() || Random::uniform() >= inter_group_prob() ) { 

            free(pin);
            continue;
         }
      }

      for ( i = 0; i < ph->vsize(); i++ ) {
         vecent *ve = ph->vecs(i);
         osrc = ve->src;
         seqno = ve->seqno;

         if ( pin->osrc == osrc && pin->seqno == seqno ) {
            should_send = true;
            break;
         }
      }

      if ( should_send ) {
         pin->nhop = dest;

         if ( cmode() == UNICAST ) {
            if ( dest == pin->dest ) {
               pin->dhtype = DTNMCTYPE_PKT;
            } else {
               pin->dhtype = DTNMCTYPE_DELEGATE;
            }
         } else if ( cmode() == MULTICAST ) {
            if ( gid == pin->gid ) {
               pin->dhtype = DTNMCTYPE_MCAST_PKT;
            } else {
               pin->dhtype = DTNMCTYPE_DELEGATE;
            }
         } else {
            pin->dhtype = DTNMCTYPE_DELEGATE;
         }

         pin->src = ra_addr();
         log("dlg node %d: nhop: %d, src: %d, osrc: %d, dest: %d, seqno: %d, nocopy: %d (time: %g)\n", ra_addr(), dest, pin->src, pin->osrc, pin->dest, pin->seqno, pin->nocopy, CURRENT_TIME, pin->ttk);
         log("LOG: e _%d_ _%d_ [%d] %g %g\n", ra_addr(), pin->nhop, pin->seqno, pin->gen_time, CURRENT_TIME);
         send_dtnmc_pkt(pin);
      }
      free(pin);
   }

   Packet::free(p);
}

void Dtnmc::recv_sumvec_pkt(Packet *p)
{
   int i;
   int count;

   struct hdr_cmn *ch = HDR_CMN(p);
   struct hdr_ip *ih = HDR_IP(p);
   struct hdr_sumvec_pkt *ph = HDR_SUMVEC_PKT(p);

   if ( ra_addr() != ph->pkt_dest() ) {
      Packet::free(p);
      return;
   }

   int *sarr = new int[VECSIZE];
   int *qarr = new int[VECSIZE];

   assert(ih->sport() == RT_PORT);
   assert(ih->dport() == RT_PORT);

   if ( ph->vsize() < 1 ) {
      Packet::free(p);
      return;
   }

   nsaddr_t sid = ph->pkt_src();

   count = 0;
   for ( i = 0; i < ph->vsize(); i++ ) {
      vecent *ve = ph->vecs(i);

      if ( rtable_.is_pkt_stored(ve->src, ve->seqno) ) { // we have this packet, no need to request
         continue;
      }

      sarr[count] = ve->src;
      qarr[count] = ve->seqno;
      count++;
   }

   send_reqvec_pkt(sid, sarr, qarr, count);

   delete sarr;
   delete qarr;

   Packet::free(p);
   return;
}

void Dtnmc::send_reqvec_pkt(nsaddr_t dest, int *sarr, int *qarr, const int count)
{
   int i;
   int n;
   Packet *p                 = allocpkt();
   struct hdr_cmn *ch        = HDR_CMN(p);
   struct hdr_ip  *ih        = HDR_IP(p);
   struct hdr_sumvec_pkt *ph = HDR_SUMVEC_PKT(p);

   ph->pkt_src()      = ra_addr();
   ph->pkt_dest()     = dest;
   ph->mcast_grp_id() = mcast_base_addr();
   ph->dh_type()      = DTNMCTYPE_REQVEC;     

   ch->ptype()        = PT_DTNMC;
   ch->direction()    = hdr_cmn::DOWN;
   ch->size()         = IP_HDR_LEN + ph->size();
   ch->error()        = 0;
   ch->next_hop()     = IP_BROADCAST;
   ch->addr_type()    = NS_AF_INET;

   ih->saddr()        = ra_addr();
   ih->daddr()        = dest;
   ih->sport()        = RT_PORT;
   ih->dport()        = RT_PORT;
   ih->ttl()          = DTNMC_DEF_TTL;

   n = 0;
   for ( i = 0; i < count; i++ ) {
      vecent *ve = ph->vecs(i);

      if ( rtable_.is_pkt_stored(ve->src, ve->seqno) == false ) { // only request packets that we don't have
         ve->src    = sarr[i];
         ve->seqno  = qarr[i];
         n++;
      }
   }

   if ( n == 0 ) {
      Packet::free(p);
      return;
   }

   ph->vsize() = n;

   log("LOG: Q %d %d %d\n", ra_addr(), dest, ph->vsize() );
   Scheduler::instance().schedule(target_, p, JITTER);
}

void Dtnmc::send_sumvec_pkt(nsaddr_t dest)
{
   int i;
   int *sarr = new int[VECSIZE];
   int *qarr = new int[VECSIZE];

   double ltime = rtable_.last_sumvecTime(dest);

   // get entries that we have received during the last SUMVEC_INTV interval
   int count = rtable_.get_stored_entries(dest, sarr, qarr, VECSIZE, ltime);

   if ( count == 0 ) {
      delete sarr;
      delete qarr;
      return;
   }

   Packet *p                 = allocpkt();
   struct hdr_cmn *ch        = HDR_CMN(p);
   struct hdr_ip  *ih        = HDR_IP(p);
   struct hdr_sumvec_pkt *ph = HDR_SUMVEC_PKT(p);

   ph->pkt_src()      = ra_addr();
   ph->pkt_dest()     = dest;
   ph->mcast_grp_id() = mcast_base_addr();
   ph->dh_type()      = DTNMCTYPE_SUMVEC;     

   ch->ptype()        = PT_DTNMC;
   ch->direction()    = hdr_cmn::DOWN;
   ch->size()         = IP_HDR_LEN + ph->size();
   ch->error()        = 0;
   ch->next_hop()     = IP_BROADCAST;
   ch->addr_type()    = NS_AF_INET;

   ih->saddr()        = ra_addr();
   ih->daddr()        = dest;
   ih->sport()        = RT_PORT;
   ih->dport()        = RT_PORT;
   ih->ttl()          = DTNMC_DEF_TTL;

   /*
    * We only update the sumvecTime if we are sure that there are no more messages
    * to be sent. 
    */
   if ( count < VECSIZE ) { 
      rtable_.update_sumvecTime(dest, CURRENT_TIME);
   }

   for ( i = 0; i < count; i++ ) {
      vecent *ve = ph->vecs(i);
      ve->src    = sarr[i];
      ve->seqno  = qarr[i];
   }
   ph->vsize() = count;
   delete sarr;
   delete qarr;

   log("LOG: v %d %d %d\n", ra_addr(), dest, ph->vsize() );
   Scheduler::instance().schedule(target_, p, JITTER);
}

void Dtnmc::send_stored_pkt(nsaddr_t dest)
{
   // see if we have any packets stored for this neighbour node
   struct dtnmc_pkt_info *pin;
   while ( (pin = rtable_.pop_packet_info(dest)) != NULL ) {
      //if ( pin->nhop != ra_addr() && pin->dest != ra_addr() ) {
      if ( pin->dest != ra_addr() ) {
         pin->nhop = dest;
         pin->dhtype = DTNMCTYPE_PKT;
         log("node: %d: sending to %d, seq_num: %d, osrc: %d (time: %g)\n", ra_addr(), pin->dest, pin->seqno, pin->osrc, CURRENT_TIME);
         // delayed send
         log("LOG: d _%d_ _%d_ [%d] %g\n", ra_addr(), pin->dest, pin->seqno, CURRENT_TIME);
         send_dtnmc_pkt(pin);
      }
      free(pin);
   }
}

void Dtnmc::send_stored_mcast_pkt(nsaddr_t dest, nsaddr_t gid)
{
   // see if we have any packets stored for this neighbour node
   struct dtnmc_pkt_info *pin;
   while ( (pin = rtable_.pop_mcast_packet_info(dest, gid)) != NULL ) {
      if ( rtable_.packet_expired(pin) == true) { // if we come across any expired packets, call to clean them
          rtable_.buffer_clean();
          free(pin);
          continue;
      }

      if ( pin->dest != ra_addr() ) {
         pin->nhop = dest;
         pin->dest = dest;
         pin->dhtype = DTNMCTYPE_MCAST_PKT;
         // delayed send
         log("LOG: D _%d_ _%d_ [%d] %g\n", ra_addr(), pin->dest, pin->seqno, CURRENT_TIME);
         send_dtnmc_pkt(pin);
      }
      free(pin);
   }
}

void Dtnmc::delegate_pkt(nsaddr_t dest)
{
   int i = 0;
   int MAX = 400; // temporary parameter. TODO: examine its effects.
   struct dtnmc_pkt_info *pin;
   //while ( (pin = rtable_.pop_delegate_packet_info(dest)) != NULL && (i++ < MAX) ) {
   while ( (pin = rtable_.pop_delegate_packet_info(dest)) != NULL ) {
      if ( pin->dest != ra_addr() ) {
         pin->nhop = dest;
         pin->dhtype = DTNMCTYPE_DELEGATE;
         pin->src = ra_addr();
         log("dlg node %d: nhop: %d, src: %d, osrc: %d, dest: %d, seqno: %d, nocopy: %d (time: %g)\n", ra_addr(), dest, pin->src, pin->osrc, pin->dest, pin->seqno, pin->nocopy, CURRENT_TIME);
         log("LOG: e _%d_ _%d_ [%d] %g %g\n", ra_addr(), pin->nhop, pin->seqno, pin->gen_time, CURRENT_TIME);
         send_dtnmc_pkt(pin);
      }
      free(pin);
   }
}

void Dtnmc::delegate_mcast_pkt(nsaddr_t dest, nsaddr_t gid)
{
   int i = 0;
   int MAX = 8; // temporary parameter. TODO: examine its effects.
   struct dtnmc_pkt_info *pin;

   // gossip support for delegation
   bool shouldSend;
   float GOSSIP_PROB = 0.8; // gossip probablity
   int   HOP_MAX = 3;  // after this many hops, we send the packet with a probablity of GOSSIP_PROB

   while ( (pin = rtable_.pop_delegate_mcast_packet_info(dest, gid)) != NULL && (i++ < MAX) ) {
      shouldSend = true;
      // spray and wait send
      //if ( pin->ttl < (DTNMC_DEF_TTL - HOP_MAX) && Random::uniform() > GOSSIP_PROB ) { 
      if ( pin->dest == ra_addr() ) {
         shouldSend = false;
      }

      //log("pin->ttl: %d\n", pin->ttl);

      if ( pin->ttl < (DTNMC_DEF_TTL - HOP_MAX) && Random::uniform() > GOSSIP_PROB ) { 
      //if ( Random::uniform() >= GOSSIP_PROB ) { 
         shouldSend = false;
      }

      if ( shouldSend == true ) {
         pin->nhop = dest;
         //pin->dest = dest;
         pin->dhtype = DTNMCTYPE_DELEGATE;
         pin->src = ra_addr();
         log("DLG node %d: nhop: %d, src: %d, osrc: %d, dest: %d, seqno: %d, nocopy: %d (time: %g)\n", ra_addr(), dest, pin->src, pin->osrc, pin->dest, pin->seqno, pin->nocopy, CURRENT_TIME);
         log("LOG: E _%d_ _%d_ [%d] [%d] %g\n", ra_addr(), pin->nhop, pin->seqno, gid, CURRENT_TIME);
         send_dtnmc_pkt(pin);
      }
      free(pin);
   }
}

void Dtnmc::delegate_to_core(nsaddr_t dest)
{
   struct dtnmc_pkt_info *pin;
   int gid = -1;

   while ( (pin = rtable_.extract_packet_info(dest)) != NULL ) {
      pin->nhop = dest;
      //pin->dest = dest;
      pin->dhtype = DTNMCTYPE_DELEGATE;
      pin->src = ra_addr();
      log("DLG node %d: nhop: %d, src: %d, osrc: %d, dest: %d, seqno: %d, nocopy: %d (time: %g)\n", ra_addr(), dest, pin->src, pin->osrc, pin->dest, pin->seqno, pin->nocopy, CURRENT_TIME);
      log("LOG: C _%d_ _%d_ [%d] [%d] %g %g\n", ra_addr(), pin->nhop, pin->seqno, gid, pin->gen_time, CURRENT_TIME);
      send_dtnmc_pkt(pin);
      free(pin);
   }
}

void Dtnmc::send_dtnmc_pkt() {
   int da;

   if ( num_nodes_ < 2 ) {
      fprintf ( stderr, "send_dtnmc_pkt(): Only one node in the system. returning...\n");
      return;
   }

   // don't send loopback packets, search till we find another node id 
   while ((da = rand() % num_nodes_) == ra_addr()) { 
      ;
   }

   seq_num_++;

   struct dtnmc_pkt_info *pin = (struct dtnmc_pkt_info *) malloc(sizeof(struct dtnmc_pkt_info));

   assert(pin != NULL);

   pin->osrc = ra_addr(); // we are the originating node
   pin->src = ra_addr();  // we are the sending node
   pin->dest = da;
   pin->nhop = da;
   pin->gid = mcast_base_addr();  // this is not a multicast packet
   pin->seqno = seq_num_; 
   pin->gen_time = CURRENT_TIME; 
   pin->member_since = CURRENT_TIME; 
   pin->member_till  = CURRENT_TIME + MEMBERSHIP_INTV; 
   pin->nocopy = NOCOPY; // number of copies to be sent for this packet
   pin->dhtype = DTNMCTYPE_PKT;
   pin->ttl = DTNMC_DEF_TTL; 
   pin->ttk = DTNMC_DEF_TTK; 

   log("LOG: g _%d_ _%d_ [%d] %g\n", ra_addr(), da, pin->seqno, CURRENT_TIME);

   send_dtnmc_pkt(pin);
   free(pin);
}

void Dtnmc::send_dtnmc_mcast_pkt() {
   if ( num_nodes_ < 2 ) {
      fprintf ( stderr, "send_dtnmc_pkt(): Only one node in the system. returning...\n");
      return;
   }

   seq_num_++;

   struct dtnmc_pkt_info *pin = (struct dtnmc_pkt_info *) malloc(sizeof(struct dtnmc_pkt_info));

   assert(pin != NULL);

   pin->osrc = ra_addr(); // we are the originating node
   pin->src = ra_addr();  // we are the sending node
   pin->dest = IP_BROADCAST;
   pin->nhop = IP_BROADCAST;
   pin->gid  = mcast_base_addr(); 
   pin->seqno = seq_num_; 
   pin->gen_time = CURRENT_TIME; 
   pin->member_since = CURRENT_TIME; 
   pin->member_till  = CURRENT_TIME + MEMBERSHIP_INTV; 
   pin->nocopy = NOCOPY; // number of copies to be sent for this packet
   pin->dhtype = DTNMCTYPE_MCAST_PKT;
   pin->ttl = DTNMC_DEF_TTL; 
   pin->ttk = DTNMC_DEF_TTK; 

   log("LOG: G _%d_ [%d] [%d] %g\n", ra_addr(), pin->seqno, pin->gid, CURRENT_TIME);

   send_dtnmc_pkt(pin);
   free(pin);
}

// send a packet to destination address da, with sequence number seqno
void Dtnmc::send_dtnmc_pkt(struct dtnmc_pkt_info *pin) {
   Packet *p                 = allocpkt();
   struct hdr_cmn *ch        = HDR_CMN(p);
   struct hdr_ip  *ih        = HDR_IP(p);
   struct hdr_dtnmc_pkt  *ph = HDR_DTNMC_PKT(p);

   ph->pkt_src()      = ra_addr();
   ph->mcast_grp_id() = pin->gid; 
   ph->pkt_dest()     = pin->dest;
   ph->pkt_osrc()     = pin->osrc;
   ph->no_copy()      = pin->nocopy;
   ph->gen_time()     = pin->gen_time;
   ph->member_since() = pin->member_since;
   ph->member_till()  = pin->member_till;
   ph->pkt_seq_num()  = pin->seqno; 
   ph->dh_type()      = pin->dhtype; 
   ph->ttk()          = pin->ttk; 

   ch->ptype()        = PT_DTNMC;
   ch->direction()    = hdr_cmn::DOWN;
   ch->size()         = IP_HDR_LEN + ph->size();
   ch->error()        = 0;
   ch->next_hop()     = IP_BROADCAST;
   ch->addr_type()    = NS_AF_INET;

   ih->saddr()        = ra_addr();
   ih->daddr()        = pin->nhop;
   ih->sport()        = RT_PORT;
   ih->dport()        = RT_PORT;
   //ih->ttl()          = DTNMC_DEF_TTL;
   ih->ttl()          = pin->ttl;

   nsaddr_t da = pin->nhop;

   if ( ph->dh_type() == DTNMCTYPE_PKT ) { // unicast packet
      // see if we have heard from the destination node recently
      double lastHeard = rtable_.last_hello(da);

      if ( CURRENT_TIME - lastHeard > HELLO_EXPIRATION_INTV ) { 
         // did not hear from destination recently, store to try later
         log("node %d: storing send request for %d at %g (last heard %g)\n", ra_addr(), ph->pkt_dest(), CURRENT_TIME, lastHeard);
         log("LOG: t _%d_ _%d_ [%d] [%d] %g\n", ra_addr(), da, pin->gid, ph->pkt_seq_num(), CURRENT_TIME);
         rtable_.store_packet(p);
         Packet::free(p);
         return;
      }

      log("LOG: s _%d_ _%d_ [%d] %g\n", ra_addr(), pin->dest, pin->seqno, CURRENT_TIME);
      Scheduler::instance().schedule(target_, p, JITTER);
   } else if ( ph->dh_type() == DTNMCTYPE_MCAST_PKT ) { // multicast packet
      if ( pin->nhop == IP_BROADCAST ) { 
         /* This multicast packet is just generated, not for a specific node.
          * We just store the multicast packets in buffer. They will be sent via delegation.
          * The main reason is to keep track of whom we have sent multicast packets to.
          * Besides, it is easier to do spray and wait this way.
          */
         nsaddr_t gid = mcast_base_addr();
         log("LOG: T _%d_ _%d_ [%d] [%d] %g\n", ra_addr(), pin->dest, pin->seqno, pin->gid, CURRENT_TIME);
         rtable_.store_packet(p);
         Packet::free(p);
      } else { // this is for a specific node
         log("LOG: S _%d_ _%d_ [%d] [%d] %g\n", ra_addr(), pin->dest, pin->seqno, pin->gid, CURRENT_TIME);
         Scheduler::instance().schedule(target_, p, JITTER);
      }
   } else if ( ph->dh_type() == DTNMCTYPE_DELEGATE ) {
      log("LOG: p _%d_ _%d_ (%d) [%d] %g\n", ra_addr(), pin->dest, pin->nhop, pin->seqno, CURRENT_TIME);
      Scheduler::instance().schedule(target_, p, JITTER);
   }

   //Scheduler::instance().schedule(target_, p, JITTER);
}

void Dtnmc::sendHello() {
   Packet *p                 = allocpkt();
   struct hdr_cmn *ch        = HDR_CMN(p);
   struct hdr_ip  *ih        = HDR_IP(p);
   struct hdr_dtnmc_pkt *ph  = HDR_DTNMC_PKT(p);  
 
   int i;

   ph->pkt_src()      = ra_addr();
   ph->mcast_grp_id() = mcast_base_addr();
   //ph->pkt_seq_num()  = seq_num_++;
   ph->pkt_seq_num()  = seq_num_;    // not increasing sequence number for hello messages
   ph->dh_type()      = DTNMCTYPE_HELLO; 
   ph->ttk()          = DTNMC_DEF_TTK; 

   if ( core_ ) {
      ph->is_core() = true;
   } else {
      ph->is_core() = false;
   }

   int *ip = rtable_.get_neighbors(MAX_NEIGHBOR);
   for ( i = 0; i < MAX_NEIGHBOR; i++ ) {
      ph->neighbor(i) = ip[i]; 
   }
   delete[] ip;

   ch->ptype()        = PT_DTNMC;
   ch->direction()    = hdr_cmn::DOWN;
   ch->size()         = IP_HDR_LEN + ph->size();
   ch->error()        = 0;
   ch->next_hop()     = IP_BROADCAST;
   ch->addr_type()    = NS_AF_INET;

   ih->saddr()        = ra_addr();
   ih->daddr()        = IP_BROADCAST;
   ih->sport()        = RT_PORT;
   ih->dport()        = RT_PORT;
   ih->ttl()          = DTNMC_DEF_TTL;

   Scheduler::instance().schedule(target_, p, SJITTER);
}

void Dtnmc::reset_pkt_timer() {
   pkt_timer_.resched((double) (PKT_INTERVAL*Random::uniform()));
}

void Dtnmc::reset_hello_timer() {

   hello_timer_.resched((double) HELLO_INTERVAL);

   rtable_.buffer_clean();

   /* To avoid small buffer occupancy report, we only report buffer occupancy after TTK time passes, 
    * and TTK time before the simulation stop time, as nodes stop generating messages at 
    * time (stop_time)
    */
   if ( msg_ttk() < CURRENT_TIME && CURRENT_TIME < stop_time() ) { 

      // log buffer occupancy at every hello interval:
      //log("LOG: b _%d_ %d %g %g\n", ra_addr(), rtable_.buffer_size(), rtable_.buffer_occupancy(), CURRENT_TIME);
      if ( ++brcount % BRFREQ == 0 ) { 
         if ( core_ ) {
            log("LOG: B _%d_ %d %g %g\n", ra_addr(), rtable_.buffer_size(), rtable_.buffer_occupancy(), CURRENT_TIME);
         } else {
            log("LOG: b _%d_ %d %g %g\n", ra_addr(), rtable_.buffer_size(), rtable_.buffer_occupancy(), CURRENT_TIME);
         }
      }
   } 

   return;
}

void Dtnmc::forward_data(Packet *p) {
   struct hdr_cmn* ch = HDR_CMN(p);
   struct hdr_ip*  ih = HDR_IP(p);

   if ( ch->direction() == hdr_cmn::UP &&
        ((u_int32_t) ih->daddr() == IP_BROADCAST || ih->daddr() == ra_addr())) {
      dmux_->recv(p, 0);
      //dmux_->recv(p, 0.0);
      return;
   } else {
      ch->direction() = hdr_cmn::DOWN;
      ch->addr_type() = NS_AF_INET;

      if ((u_int32_t) ih->daddr() == IP_BROADCAST) {
         ch->next_hop() = IP_BROADCAST;
      } else {
         nsaddr_t next_hop = rtable_.lookup(ih->daddr());
         if ( next_hop == IP_BROADCAST ) {
            debug("%f: Agent %d can not forward a packet destined to %d\n",
                  CURRENT_TIME, ra_addr(), ih->daddr());
            drop(p, DROP_RTR_NO_ROUTE);
            return;
         } else {
            ch->next_hop() = next_hop;
         }
      }

      Scheduler::instance().schedule(target_, p, 0.0);
   }
}

/* log(const char *fmt, ...) -- logs messages given in variable length arguments.
 * Currently using stderr, mainly because there does not seem to be an easy way
 * to tell the file name to the agent (which is written in C++) by the OTCL script.
 * The bind family of methods does not bind strings. Setting by calling a function of
 * the agent through OTCL script has not been tried yet.
 */
void Dtnmc::log(const char *fmt, ...)
{
   FILE *fp = stderr; // TODO: change this to a regular file
 
   char buf[MAXLINE];

   va_list         ap;

   va_start(ap, fmt);

   vsnprintf(buf, sizeof(buf), fmt, ap);   
   fflush(fp);
   fprintf(fp, "%s", buf);

   va_end(ap);
   return;
}
