#ifndef __dtnmc_h__
#define __dtnmc_h__

#include "dtnmc_pkt.h"
#include "dtnmc_rtable.h"

#include <agent.h>
#include <packet.h>
#include <trace.h>
#include <timer-handler.h>
#include <random.h>
#include <classifier-port.h>

#define CURRENT_TIME        (Scheduler::instance().clock())
#define JITTER              (Random::uniform()*0.5)
#define SJITTER             (Random::uniform()*0.01) // small jitter

//#define PKT_INTERVAL (100.0*1000)
#define PKT_INTERVAL 2000.0
#define HELLO_INTERVAL 3.0
#define HELLO_EXPIRATION_INTV (HELLO_INTERVAL + 0.75) 
#define MAXLINE 1024

#define NOCOPY (Dtnmc::nocopy())

#define SUMVEC_INTV 3.0

// Reporting buffer occupancy at every hello interval produces huge amounts of data. For this, we use buffer
// report frequency (actually, the inverse: report buffer occupancy at every BRFREQ hello intervals).
#define BRFREQ    50 

#define UNICAST   0x01
#define MULTICAST 0x02

#define MEMBERSHIP_INTV (Dtnmc::msg_ttk())
#define DTNMC_DEF_TTL (Dtnmc::msg_ttl())
#define DTNMC_DEF_TTK (Dtnmc::msg_ttk())

class Dtnmc;

/* Timers */
class Dtnmc_PktTimer : public TimerHandler {
public:
   Dtnmc_PktTimer ( Dtnmc* agent ) : TimerHandler() {
      agent_ = agent;
   }

protected:
   Dtnmc* agent_;
   virtual void expire (Event *e);
};

class Dtnmc_HelloTimer : public TimerHandler {
public:
   Dtnmc_HelloTimer ( Dtnmc* agent ) : TimerHandler() {
      agent_ = agent;
   }

protected:
   Dtnmc* agent_;
   virtual void expire (Event *e);
};

/* Agent */

class Dtnmc : public Agent {

   /* Friends */
   friend class Dtnmc_PktTimer;
   friend class Dtnmc_HelloTimer;

   /* private members */
   nsaddr_t      ra_addr_;
   nsaddr_t      mcast_base_addr_;
   MobileNode* thisnode;
   //dtnmc_state   state_;
   dtnmc_rtable  rtable_;
   int           spray_;
   int           nocopy_;
   int           infect_;
   int           core_;         // whether node is core (supernode)
   int           nstat_;        // whether the node is static node (currently for statistically purposes only)
   int           xpos_;         // position of node on x axis (for static nodes) 
   int           ypos_;         // position of node on y axis (for static nodes) 
   int           num_nodes_;    // number of nodes in the system
   int           seq_num_;      // message sequence numbers
   double        *lastHello;    // record Hello messages for previous contacts
   double        member_since_; // group member since
   double        member_till_;  // group member till
   int           cmode_;        // communication mode: currently either multicasting or unicasting
   double        stop_time_;    // time after which nodes do not generate new messages (receive only mode)
   double        msg_ttk_;      // time to kill
   int           msg_ttl_;      // time to live (number of hops allowed for messages)
   int           brcount;       // buffer report counter to control the frequncy of buffer occupancy reporting
   int           intra_group_;  // indicates if messages are communicated within group only 
   double        inter_group_prob_;  // indicates probability with which nodes exchange packets with non-group members 

protected:
   PortClassifier*  dmux_;           // For passing packets up to agents.
   Trace*           logtarget_;      // For logging.
   Dtnmc_PktTimer   pkt_timer_;      // Timer for sending packets.
   Dtnmc_HelloTimer hello_timer_;    // Timer for sending hello packets.

   inline int&          spray()  { return spray_; }
   inline int&          nocopy()  { return nocopy_; }
   inline nsaddr_t&     ra_addr()         { return ra_addr_; }
   inline nsaddr_t&     mcast_base_addr() { return mcast_base_addr_; }
   inline double&       member_since() { return member_since_; }
   inline double&       member_till()  { return member_till_; }
   inline double&       stop_time()  { return stop_time_; }
   inline double&       msg_ttk()  { return msg_ttk_; }
   inline int&          msg_ttl()  { return msg_ttl_; }
   inline int&          intra_group()  { return intra_group_; }
   inline double&       inter_group_prob()  { return inter_group_prob_; }
   inline int&          cmode()  { return cmode_; }
   inline int&          core()  { return core_; }
   inline int&          nstat()  { return nstat_; }
   inline int&          xpos()  { return xpos_; }
   inline int&          ypos()  { return ypos_; }

   void forward_data(Packet *p);
   void recv_dtnmc_pkt(Packet *p);
   void recv_dtnmc_mcast_pkt(Packet *p);
   void recv_delegate_pkt(Packet *p);
   void recv_hello_pkt(Packet *p);
   void send_dtnmc_pkt();
   void send_dtnmc_mcast_pkt();
   void send_dtnmc_pkt(struct dtnmc_pkt_info *pin);
   void send_sumvec_pkt(nsaddr_t dest);
   void recv_sumvec_pkt(Packet*);
   void send_reqvec_pkt(nsaddr_t dest, int *sarr, int *qarr, const int count);
   void recv_reqvec_pkt(Packet*);
   void send_stored_pkt(nsaddr_t dest);
   void send_stored_mcast_pkt(nsaddr_t dest, nsaddr_t gid);
   void delegate_pkt(nsaddr_t dest);
   void delegate_to_core(nsaddr_t);
   void delegate_mcast_pkt(nsaddr_t dest, nsaddr_t gid);
   void sendHello();
   void log(const char *fmt, ...);

   void reset_pkt_timer();
   void reset_hello_timer();

public:
   Dtnmc(nsaddr_t id); // node id and total number of nodes in the system
   int  command(int, const char *const*);
   void recv(Packet *p, Handler *h);
};

#endif
