#include <trace/cmu-trace.h>
#include <routing/address.h>
#include <string>

using std::string;

#include "dtnmc_rtable.h"
#include "dtnmc.h"
#include "dtnmc_pkt.h"

#define CURRENT_TIME        (Scheduler::instance().clock())
//#define DMIN (-10000000)

dtnmc_rtable::dtnmc_rtable()
{
   // default to BUF_SIZE if no parameter is given for capacity
   // dtnmc_rtable(BUF_SIZE, APP_BUF_SIZE);
   capacity = BUF_SIZE; // capacity of packet buffer (pinfo_)
   app_capacity = APP_BUF_SIZE;
   source_spray_ = false; // by default we use Binary Spray and Wait
   epidemic_ = false; // by default epidemic is turned off 
}

dtnmc_rtable::dtnmc_rtable(u_int32_t cap)
{
   //dtnmc_rtable(cap, APP_BUF_SIZE);
   capacity = cap; // capacity of packet buffer (pinfo_)
   app_capacity = APP_BUF_SIZE;
   source_spray_ = false; // by default we use Binary Spray and Wait
   epidemic_ = false; // by default epidemic is turned off 
}

dtnmc_rtable::dtnmc_rtable(u_int32_t cap, u_int32_t appcap)
{
   capacity = cap; // capacity of packet buffer (pinfo_)
   app_capacity = appcap;
   source_spray_ = false; // by default we use Binary Spray and Wait
   epidemic_ = false; // by default epidemic is turned off 
}

void dtnmc_rtable::set_capacity(int cap)
{
   assert(cap>0);
   capacity = cap;
}

void dtnmc_rtable::print(Trace *out) {
   sprintf(out->pt_->buffer(), "P\tdest\tnext");
   out->pt_->dump();
   for(rtable_t::iterator it = rt_.begin(); it != rt_.end(); it++) {
      sprintf(out->pt_->buffer(), "P\t%d\t%d", (*it).first, (*it).second);
      out->pt_->dump();
   }
}

void dtnmc_rtable::clear() {
   rt_.clear();
}

void dtnmc_rtable::rm_entry(nsaddr_t dest) {
   rt_.erase(dest);
}

void dtnmc_rtable::add_entry(nsaddr_t dest, nsaddr_t next) {
   rt_[dest] = next;
}

nsaddr_t dtnmc_rtable::lookup(nsaddr_t dest) {
   rtable_t::iterator it = rt_.find(dest);

   if ( it == rt_.end() ) {
      return IP_BROADCAST;
   } else {
      return (*it).second;
   }
}

u_int32_t dtnmc_rtable::size() {
   return rt_.size();
}

void dtnmc_rtable::received_add_entry(nsaddr_t osrc, int seqno)
{
   string str = joinToString(osrc, seqno); 
   received_[str] = CURRENT_TIME;
   return;
}

int dtnmc_rtable::get_received_entries(int *sarr, int *qarr, int max, double time) {
   nseq_pair_t::iterator it;
   int count = 0;
   int s, q; 

   for(it = received_.begin(); it != received_.end(); it++) {
      if ( time < (*it).second ) {
         sscanf( ((*it).first).c_str(), "%d:%d", &s, &q);
         sarr[count] = s;
         qarr[count] = q;
         count++;

         if ( count >= max ) {
            break;
         }
      }
   }

   return count;
}

bool dtnmc_rtable::lookup_received_entry(nsaddr_t osrc, int seqno)
{
   string str = joinToString(osrc, seqno); 
   nseq_pair_t::iterator it = received_.find(str);

   if ( it == received_.end() ) {
       return false;
   }

   return true;
}

void dtnmc_rtable::delegated_add_entry(nsaddr_t osrc, nsaddr_t nhop, int seqno)
{
   string str = joinToString3(osrc, nhop, seqno); 
   delegated_[str] = CURRENT_TIME;
   return;
}

bool dtnmc_rtable::delegated_lookup_entry(nsaddr_t osrc, nsaddr_t nhop, int seqno)
{
   string str = joinToString3(osrc, nhop, seqno); 
   nseq_pair_t::iterator it = delegated_.find(str);

   if ( it == delegated_.end() ) {
       return false;
   }

   return true;
}

string dtnmc_rtable::joinToString(nsaddr_t nodeid, int seqno)
{
   char buff[30];
   sprintf(buff, "%d:%d", nodeid, seqno);
   string str(buff);
   return str;
}

string dtnmc_rtable::joinToString3(nsaddr_t osrc, nsaddr_t nodeid, int seqno)
{
   char buff[50];
   sprintf(buff, "%d:%d:%d", osrc, nodeid, seqno);
   string str(buff);
   return str;
}

void dtnmc_rtable::update_hello(nsaddr_t src, double dt) {
   hello_[src] = dt;
}

void dtnmc_rtable::update_first_hello(nsaddr_t src, double dt) {
   first_hello_[src] = dt;
}

void dtnmc_rtable::update_sumvecTime(nsaddr_t src, double dt) {
   sumvecTime_[src] = dt;
}

double dtnmc_rtable::last_hello(nsaddr_t src) {
   hello_t::iterator it = hello_.find(src);

   if ( it == hello_.end() ) { // no hello message recorded for src node
      return DMIN ;
   } else {
      return (*it).second;
   }
}

double dtnmc_rtable::first_hello(nsaddr_t src) {
   hello_t::iterator it = first_hello_.find(src);

   if ( it == first_hello_.end() ) { // no hello message recorded for src node
      return DMIN ;
   } else {
      return (*it).second;
   }
}

double dtnmc_rtable::last_sumvecTime(nsaddr_t src) {
   hello_t::iterator it = sumvecTime_.find(src);

   if ( it == sumvecTime_.end() ) { // no hello message recorded for src node
      return DMIN ;
   } else {
      return (*it).second;
   }
}

int *dtnmc_rtable::get_neighbors(int max) {
   assert(max > 0);

   int *ip = new int[max];
   int i;

   // clear out the neighbor array
   for ( i = 0; i < max; i++ ) {
      ip[i] = NEIGHBOR_END;
   }

   hello_t::iterator it;

   i = 0;
   for ( it = hello_.begin(); (i < max) && (it != hello_.end()); it++ ) {
      int    nid = (*it).first;
      double lastHeard = (*it).second;

      if ( lastHeard > (CURRENT_TIME - lastHeard > HELLO_EXPIRATION_INTV) ) {
         ip[i++] = nid; 
      }
   }

   return ip; // should be "delete"d by caller
}

void dtnmc_rtable::set_neighbors(int src, std::vector <int> new_neighbors)
{
   // clean up old record
   neighbor_t::iterator it = neighbor_.find(src);
   if ( it != neighbor_.end() ) {
      neighbor_.erase(it);
   }

   neighbor_[src] = new_neighbors;
}

void dtnmc_rtable::update_grp_hello(nsaddr_t gid, double dt) {
   grp_hello_[gid] = dt;
}

double dtnmc_rtable::last_grp_hello(nsaddr_t gid) {
   hello_t::iterator it = grp_hello_.find(gid);

   if ( it == grp_hello_.end() ) { // no hello message recorded for group ``gid''
      return DMIN ;
   } else {
      return (*it).second;
   }
}

u_int32_t dtnmc_rtable::hello_size() {
   return hello_.size();
}

void dtnmc_rtable::store_packet(Packet *p) {
   struct hdr_ip *ih = HDR_IP(p);
   struct hdr_dtnmc_pkt *ph = HDR_DTNMC_PKT(p);

   struct dtnmc_pkt_info *pin = (struct dtnmc_pkt_info *) malloc(sizeof(struct dtnmc_pkt_info));

   if (buffer_size() == capacity) { // the buffer is full
      buffer_clean(); // try to remove expired items first

      if ( buffer_size() == capacity ) { // if there are no expired item to clear, remove the oldest one
         buffer_prune();
      }
   }

   pin->nhop = ih->daddr();
   pin->ttl = ih->ttl();
   pin->dest = ph->pkt_dest(); 
   pin->dhtype = ph->dh_type(); 
   pin->src = ph->pkt_src(); 
   pin->gid = ph->mcast_grp_id(); 
   pin->osrc = ph->pkt_osrc(); 
   pin->nocopy = ph->no_copy(); 
   pin->seqno  = ph->pkt_seq_num();
   pin->gen_time = ph->gen_time();
   pin->member_since = ph->member_since();
   pin->member_till = ph->member_till();
   pin->ttk = ph->ttk();
   pin->time_stored = CURRENT_TIME;
   pin->dlg_cores = NULL; 

   pinfo_.push_back(pin);
}

int dtnmc_rtable::get_stored_entries(nsaddr_t dest, int *sarr, int *qarr, int max, double time) {
   int count = 0;

   std::vector<dtnmc_pkt_info *>::iterator it;
   for(it = pinfo_.begin(); it != pinfo_.end(); it++) {
      dtnmc_pkt_info *pi = *it;

      if ( pi->ttl < 1 ) {  // TTL has reached, will not forward
         continue;
      }

      if ( time < pi->time_stored ) {
         if(delegated_lookup_entry(pi->osrc, dest, pi->seqno)) { 
            continue;
         }

         sarr[count] = pi->osrc;
         qarr[count] = pi->seqno;

         count++;

         if ( count >= max ) {
            break;
         }
      }
   }

   return count;
}

/* buffer_clean(): delete packets that have expired.
 */
void dtnmc_rtable::buffer_clean()
{
   bool done = false;

   if ( buffer_size() == 0 ) {
      return;
   }
   
   /* I could find a better way to sequentially remove expired packets. Deleting through iteraters
    * one after another messes things up. TODO: try to do it sequentially, rather than through double loop 
    */
   int i = 0;
   while ( done == false ) {
      done = true;

      std::vector<dtnmc_pkt_info *>::iterator it;
      for(it = pinfo_.begin(); it != pinfo_.end(); it++)
      {
         dtnmc_pkt_info *pi = *it;
         if ( packet_expired(pi) ) {
             pinfo_.erase(it);
             fprintf(stderr, "node %d: dropping pi: pi->osrc: %d, pi->dest: %d, pi->time_stored: %g (%g--%g), time: %g\n", id(), pi->osrc, pi->dest, pi->time_stored, pi->member_since, pi->member_till, CURRENT_TIME);
             free(pi);
             done = false; // found one, but there might be more
             break;
         }
      }
      // if no packets found to be expired, "done" will remain true, and we will exit the loop.
   }
}

/*
 * buffer_prune() -- remove the oldest message from buffer. 
 * This operation will decrease the size of buffer by one, insha-Allah.
 */
void dtnmc_rtable::buffer_prune()
{
   dtnmc_pkt_info *pi = NULL;
   double oldest_time = CURRENT_TIME;
   int idx = 0;

   if ( buffer_size() == 0 ) {
      return;
   }

   int i = 0;
   std::vector<dtnmc_pkt_info *>::iterator it;
   for(it = pinfo_.begin(); it != pinfo_.end(); it++) {
      dtnmc_pkt_info *pi = *it;
      if ( pi->time_stored < oldest_time ) {
         oldest_time = pi->time_stored;
         idx = i;
      }

      i++;
   }

   it = pinfo_.begin();
   pinfo_.erase(it+idx);

   if ( pi != NULL ) {
      free(pi);
   }
}

/*
 * TODO: change the method to accomodate the semantic models for multicasting. 
 */
inline bool dtnmc_rtable::packet_expired ( dtnmc_pkt_info *pi )
{
   if ( pi->member_till <= CURRENT_TIME ) {
      return true;
   }

   return false;
}


inline u_int32_t dtnmc_rtable::buffer_size()
{
   return pinfo_.size();
}

double dtnmc_rtable::buffer_occupancy()
{
   return ((double)pinfo_.size())/capacity;
}

/* 
 * pop a packet stored for dest node
 */
dtnmc_pkt_info *dtnmc_rtable::pop_packet_info(nsaddr_t dest) {
   std::vector<dtnmc_pkt_info *>::iterator it;
   for(it = pinfo_.begin(); it != pinfo_.end(); it++) {
      dtnmc_pkt_info *pi = *it;

      if ( pi->dest == dest ) {
          pinfo_.erase(it);
          return pi;
      }
   }

   return NULL;
}

/* 
 * pop a packet stored for dest node
 */
bool dtnmc_rtable::is_pkt_stored(nsaddr_t src, int seqno) {
   std::vector<dtnmc_pkt_info *>::iterator it;

   for(it = pinfo_.begin(); it != pinfo_.end(); it++) {
      dtnmc_pkt_info *pi = *it;

      if ( pi->osrc == src && pi->seqno == seqno ) {
          return true;
      }
   }

   return false;
}

/* 
 * pop a packet stored for group member 'dest' with group id 'gid'
 */
dtnmc_pkt_info *dtnmc_rtable::pop_mcast_packet_info(nsaddr_t dest, nsaddr_t gid) {
   std::vector<dtnmc_pkt_info *>::iterator it;
   for(it = pinfo_.begin(); it != pinfo_.end(); it++) {
      dtnmc_pkt_info *pi = *it;

      //if ( pi->dhtype == DTNMCTYPE_MCAST_PKT && pi->gid == gid ) {
      if ( pi->gid == gid && pi->osrc != dest ) {
         if(!delegated_lookup_entry(pi->osrc, dest, pi->seqno)) { 
            delegated_add_entry(pi->osrc, dest, pi->seqno);

            struct dtnmc_pkt_info *npi = (struct dtnmc_pkt_info *) malloc(sizeof(struct dtnmc_pkt_info));

            assert(npi != NULL);

            npi->dest = pi->dest;
            npi->nhop = pi->nhop;
            npi->src = pi->src;
            npi->gid = pi->gid;
            npi->ttl = pi->ttl;
            npi->ttk = pi->ttk;
            npi->osrc = pi->osrc;
            npi->seqno  = pi->seqno;
            npi->gen_time = pi->gen_time;
            npi->member_since = pi->member_since;
            npi->member_till = pi->member_till;
            npi->time_stored =pi->time_stored;
             
            return npi;
         }
      }
   }

   return NULL;
}

/* 
 * Pop a packet info for delegation for the core. The packet info would be deleted in earlier
 * implementations. However, since there is no real need to delete the local copy, we decided
 * to keep it for better performance.
 */
dtnmc_pkt_info *dtnmc_rtable::extract_packet_info(nsaddr_t core_id) {
   /*
   if ( pinfo_.size() < capacity / 3 ) {
      return NULL;
   }
   */

   std::vector<dtnmc_pkt_info *>::iterator it;
   struct cores *carr;
   bool already_delegated;

   for(it = pinfo_.begin(); it != pinfo_.end(); it++) {
      dtnmc_pkt_info *pi = *it;

      already_delegated = false;

      for ( carr = pi->dlg_cores; carr != NULL; carr = carr->next ) {
         if ( carr->core_id == core_id ) { // we have already delegated this packet to this core
            already_delegated = true;
            break;
         }
      }

      if ( already_delegated == true ) {
         continue;
      }

      struct cores *ncore = (struct cores *) malloc ( sizeof(struct cores));
      assert(ncore);
      ncore->core_id = core_id;
      ncore->next = pi->dlg_cores;
      pi->dlg_cores = ncore;

      struct dtnmc_pkt_info *npi = (struct dtnmc_pkt_info *) malloc(sizeof(struct dtnmc_pkt_info));
      assert(npi != NULL);

      npi->dest = pi->dest;
      npi->nhop = pi->nhop;
      npi->src = pi->src;
      npi->gid = pi->gid;
      npi->ttl = pi->ttl;
      npi->ttk = pi->ttk;
      npi->osrc = pi->osrc;
      npi->seqno  = pi->seqno;
      npi->gen_time = pi->gen_time;
      npi->member_since = pi->member_since;
      npi->member_till = pi->member_till;
      npi->time_stored =pi->time_stored;

      //pinfo_.erase(it);
   
      return npi;
   }

   return NULL;
}

/* We try to implement the "binary spray and wait".
 * Pop a packet to be delegated to dest. So we don't pop packets that came from dest.
 * However, dest may receive something it sent if it came to us indirectly.
 */
dtnmc_pkt_info *dtnmc_rtable::pop_delegate_packet_info(nsaddr_t dest) {
   std::vector<dtnmc_pkt_info *>::iterator it;

   for(it = pinfo_.begin(); it != pinfo_.end(); it++) {
      dtnmc_pkt_info *pi = *it;

      // we have delegated this packet to dest, don't do it again
      if(delegated_lookup_entry(pi->osrc, dest, pi->seqno)) { 
         continue;
      }

      // don't delegate to original or previous senders
      // also, we don't return packet if number of copies (nocopy) field is 1 (see spray and wait).
      if ( pi->dest != dest && pi->src != dest && pi->osrc != dest && (epidemic() || pi->nocopy > 1) ) {
          struct dtnmc_pkt_info *npi = (struct dtnmc_pkt_info *) malloc(sizeof(struct dtnmc_pkt_info));

          assert(npi != NULL);

          npi->dest = pi->dest;
          npi->nhop = pi->nhop;
          npi->src = pi->src;
          npi->gid = pi->gid;
          npi->ttl = pi->ttl;
          npi->ttk = pi->ttk;
          npi->osrc = pi->osrc;
          npi->seqno  = pi->seqno;
          npi->gen_time = pi->gen_time;
          npi->member_since = pi->member_since;
          npi->member_till = pi->member_till;
          npi->time_stored =pi->time_stored;

          int nc;
          if ( epidemic() == true ) {
             // as long as the current copy is larger than 1, we just keep propagating it as epidemic 
             npi->nocopy = pi->nocopy;
          } else {
              // update number of copies (according to binary spray & wait, 
              // which is the default, or source spray and wait)
             if ( source_spray() == true ) {
                nc = 1; // in source spray and wait, only the source gives out copies one by one
             } else {
                // in binary spray and wait, the source gives floor(nocopy/2) copies to delegates, 
                // and delegates can further spray the message
                nc = pi->nocopy / 2; 
             }
             npi->nocopy = nc; 
             pi->nocopy  = pi->nocopy - nc;
          }

          delegated_add_entry(pi->osrc, dest, pi->seqno);

          return npi;
      }
   }

   return NULL;
}

/* We try to implement the "binary spray and wait".
 * Pop a packet to be delegated to dest. So we don't pop packets that came from dest.
 * However, dest may receive something it sent if it came to us indirectly.
 */
dtnmc_pkt_info *dtnmc_rtable::pop_delegate_mcast_packet_info(nsaddr_t dest, nsaddr_t gid) {
   std::vector<dtnmc_pkt_info *>::iterator it;

   for(it = pinfo_.begin(); it != pinfo_.end(); it++) {
      dtnmc_pkt_info *pi = *it;

      // we have delegated this packet to dest, don't do it again
      if(delegated_lookup_entry(pi->osrc, dest, pi->seqno)) { 
         continue;
      }

      // don't delegate to original or previous senders
      // also, we don't return packet if number of copies (nocopy) field is 1 (see spray and wait).
      if ( pi->dest != dest && pi->src != dest && pi->osrc != dest && pi->nocopy > 1 ) {
          struct dtnmc_pkt_info *npi = (struct dtnmc_pkt_info *) malloc(sizeof(struct dtnmc_pkt_info));

          assert(npi != NULL);

          if ( pi->ttl > 0 ) {
             pi->ttl = pi->ttl - 1;
          }

          npi->dest = pi->dest;
          npi->nhop = pi->nhop;
          npi->ttl = pi->ttl;
          npi->ttk = pi->ttk;
          npi->src = pi->src;
          npi->gid = pi->gid;
          npi->osrc = pi->osrc;
          npi->seqno  = pi->seqno;
          npi->gen_time = pi->gen_time;
          npi->member_since = pi->member_since;
          npi->member_till = pi->member_till;
          npi->time_stored =pi->time_stored;

          // update number of copies (according to binary spray & wait)
          int nc = pi->nocopy / 2;
          npi->nocopy = nc; 
          pi->nocopy  = pi->nocopy - nc;

          delegated_add_entry(pi->osrc, dest, pi->seqno);

          return npi;
      }
   }

   return NULL;
}

void dtnmc_rtable::push_packet_info(nsaddr_t dest, int seqno) {
}

void dtnmc_rtable::store_app_packet(Packet *p) {
   return ;
   struct hdr_ip *ih = HDR_IP(p);
   struct hdr_dtnmc_pkt *ph = HDR_DTNMC_PKT(p);

   struct dtnmc_pkt_info *pin = (struct dtnmc_pkt_info *) malloc(sizeof(struct dtnmc_pkt_info));

   if (app_buff_.size() == app_capacity) { // the buffer is full
      app_buffer_clean(); // try to remove expired items first

      if ( app_buff_.size() == app_capacity ) { //if there're no expired items to clear, remove the oldest one
         app_buffer_prune();
      }
   }

   pin->nhop = ih->daddr();
   pin->ttl = ih->ttl();
   pin->dest = ph->pkt_dest(); 
   pin->dhtype = ph->dh_type(); 
   pin->src = ph->pkt_src(); 
   pin->gid = ph->mcast_grp_id(); 
   pin->osrc = ph->pkt_osrc(); 
   pin->nocopy = ph->no_copy(); 
   pin->seqno  = ph->pkt_seq_num();
   pin->gen_time = ph->gen_time();
   pin->ttk = ph->ttk();
   pin->member_since = ph->member_since();
   pin->member_till = ph->member_till();
   pin->time_stored = CURRENT_TIME;
   pin->dlg_cores = NULL;

   app_buff_.push_back(pin);
}

/* app_buffer_clean(): delete packets that have expired.
 */
void dtnmc_rtable::app_buffer_clean()
{
   bool done = false;

   if ( app_buff_.size() == 0 ) {
      return;
   }
   
   /* I could not find a better way to sequentially remove expired packets. Deleting through iterators
    * one after another messes things up. TODO: try to do it sequentially, rather than through double loop 
    */
   int i = 0;
   while ( done == false ) {
      done = true;

      std::vector<dtnmc_pkt_info *>::iterator it;
      for(it = app_buff_.begin(); it != app_buff_.end(); it++)
      {
         dtnmc_pkt_info *pi = *it;
         if ( packet_expired(pi) ) {
             app_buff_.erase(it);
             fprintf(stderr, "node %d: dropping appi: pi->osrc: %d, pi->dest: %d, pi->time_stored: %g, time: %g\n", id(), pi->osrc, pi->dest, pi->time_stored, CURRENT_TIME);
             free(pi);
             done = false; // found one, but there might be more
             break;
         }
      }
      // if no packets found to be expired, "done" will remain true, and we will exit the loop.
   }
}

/*
 * app_buffer_prune() -- remove the oldest message from application buffer. 
 * This operation will decrease the size of buffer by one, insha-Allah.
 */
void dtnmc_rtable::app_buffer_prune()
{
   dtnmc_pkt_info *pi = NULL;
   double oldest_time = CURRENT_TIME;
   int idx = 0;

   if ( app_buff_.size() == 0 ) {
      return;
   }

   int i = 0;
   std::vector<dtnmc_pkt_info *>::iterator it;
   for(it = app_buff_.begin(); it != app_buff_.end(); it++) {
      dtnmc_pkt_info *pi = *it;
      if ( pi->time_stored < oldest_time ) {
         oldest_time = pi->time_stored;
         idx = i;
      }

      i++;
   }

   it = app_buff_.begin();
   app_buff_.erase(it+idx);

   if ( pi != NULL ) {
      free(pi);
   }
}

inline bool dtnmc_rtable::app_packet_expired ( dtnmc_pkt_info *pi )
{
   if ( pi->member_till < CURRENT_TIME ) {
      return true;
   }

   return false;
}
