#ifndef __dtnmc_rtable_h__
#define __dtnmc_rtable_h__

#include <trace.h>
#include <map.h>
#include <vector>
#include <string>

#include <packet.h> 
#include <routing/address.h> 

#include "dtnmc_pkt.h"

#define  BUF_SIZE      2000   // max number of messages to be stored
#define  CORE_BUF_SIZE 10000 // max number of messages to be stored for core nodes
#define  APP_BUF_SIZE  900   // max number of messages to be stored in application buffer (phony)
#define  DMIN (-10000000)

typedef std::map<nsaddr_t, nsaddr_t> rtable_t;
typedef std::map<string, double> nseq_pair_t; // (node, seq_num) pair
typedef std::map<nsaddr_t, double> hello_t;
typedef std::map<nsaddr_t, std::vector <int> > neighbor_t;

class dtnmc_rtable {
   nsaddr_t id_;
   rtable_t rt_;
   hello_t hello_;
   hello_t first_hello_;
   hello_t sumvecTime_;
   neighbor_t neighbor_;
   hello_t grp_hello_;
   nseq_pair_t received_;
   nseq_pair_t delegated_;
   std:: vector <dtnmc_pkt_info *> pinfo_ ;
   std:: vector <dtnmc_pkt_info *> app_buff_ ;
   u_int32_t capacity;
   u_int32_t app_capacity;

   /* Default spray and wait uses binary spray, set this to true if Source Spray and Wait
    * algorithm is to be used. (Refer to "Spray and Wait: An Efficient Routing Scheme for Intermittently 
    * Connected Mobile Networks", by Spyropoulos et. al.)
    */
   bool source_spray_;  
   bool epidemic_;  

private:
   string joinToString(nsaddr_t, int);
   string joinToString3(nsaddr_t, nsaddr_t, int);

public:
   dtnmc_rtable();
   dtnmc_rtable(u_int32_t cap);
   dtnmc_rtable(u_int32_t cap, u_int32_t appcap);
   void print(Trace *);

   inline nsaddr_t& id() { return id_; }
   inline bool& source_spray() { return source_spray_; }
   inline bool& epidemic() { return epidemic_; }

   // for routing table (neighbour) entries
   void clear();
   void set_capacity(int cap);
   void rm_entry(nsaddr_t);
   void add_entry(nsaddr_t, nsaddr_t);
   nsaddr_t lookup(nsaddr_t);
   u_int32_t size();

   // for received messages
   void received_add_entry(nsaddr_t, int);
   bool lookup_received_entry(nsaddr_t, int);

   // for delegated messages
   void delegated_add_entry(nsaddr_t, nsaddr_t, int);
   bool delegated_lookup_entry(nsaddr_t, nsaddr_t, int);

   // for HELLO messages
   void   update_hello (nsaddr_t src, double ht);
   void   update_first_hello (nsaddr_t src, double ht);
   void   update_sumvecTime (nsaddr_t src, double svt);
   void   update_grp_hello (nsaddr_t gid, double ht);
   void   clear_hello();
   double last_hello (nsaddr_t src);
   double first_hello (nsaddr_t src);
   double last_sumvecTime (nsaddr_t src);
   double last_grp_hello (nsaddr_t gid);
   u_int32_t hello_size();

   // for packet buffer
   void store_packet(Packet *p);
   dtnmc_pkt_info *pop_packet_info(nsaddr_t dest);
   dtnmc_pkt_info *pop_mcast_packet_info(nsaddr_t dest, nsaddr_t gid);
   dtnmc_pkt_info *pop_delegate_packet_info(nsaddr_t dest);
   dtnmc_pkt_info *pop_delegate_mcast_packet_info(nsaddr_t dest, nsaddr_t gid);
   dtnmc_pkt_info *extract_packet_info(nsaddr_t core_id);
   void push_packet_info(nsaddr_t dest, int seqno);
   int get_received_entries(int *sarr, int *qarr, int max, double time);
   int get_stored_entries(nsaddr_t dest, int *sarr, int *qarr, int max, double time);
   u_int32_t buffer_size();
   double buffer_occupancy();
   void buffer_clean();
   void buffer_prune();
   bool packet_expired(dtnmc_pkt_info*);
   bool is_pkt_stored(nsaddr_t src, int seqno);

   // neighbor management
   int *get_neighbors(int max);
   void set_neighbors(int dest, std::vector <int> new_neighbors);

   // for application level packet buffer
   void store_app_packet(Packet *p);
   void app_buffer_prune();
   void app_buffer_clean();
   bool app_packet_expired(dtnmc_pkt_info*);
};

#endif
